package com.sun.jndi.ldap.pool;

import com.sun.jndi.ldap.LdapPoolManager;
import java.util.Iterator;
import javax.naming.CommunicationException;
import javax.naming.InterruptedNamingException;
import javax.naming.NamingException;
import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.lang.ref.Reference;
import java.util.List;

final class Connections implements PoolCallback
{
    private static final boolean debug;
    private static final boolean trace;
    private static final int DEFAULT_SIZE = 10;
    private final int maxSize;
    private final int prefSize;
    private final List<ConnectionDesc> conns;
    private boolean closed;
    private Reference<Object> ref;
    
    Connections(Object referent, int min, int n, int b, PooledConnectionFactory pooledConnectionFactory) throws NamingException {
        this.closed = false;
        this.maxSize = b;
        if (b > 0) {
            this.prefSize = Math.min(n, b);
            min = Math.min(min, b);
        }
        else {
            this.prefSize = n;
        }
        this.conns = new ArrayList((b > 0) ? b : 10);
        this.ref = new SoftReference(referent);
        this.d("init size=", min);
        this.d("max size=", b);
        this.d("preferred size=", n);
        for (int i = 0; i < min; ++i) {
            PooledConnection pooledConnection = pooledConnectionFactory.createPooledConnection(this);
            this.td("Create ", pooledConnection, pooledConnectionFactory);
            this.conns.add(new ConnectionDesc(pooledConnection));
        }
    }
    
    synchronized PooledConnection get(long n, PooledConnectionFactory pooledConnectionFactory) throws NamingException {
        long n2 = (n > 0L) ? System.currentTimeMillis() : 0L;
        this.d("get(): before");
        PooledConnection orCreateConnection;
        while ((orCreateConnection = this.getOrCreateConnection(pooledConnectionFactory)) == null) {
            if (n > 0L && n <= 0L) {
                throw new CommunicationException("Timeout exceeded while waiting for a connection: " + n + "ms");
            }
            try {
                this.d("get(): waiting");
                if (n > 0L) {
                    this.wait(n);
                }
                else {
                    this.wait();
                }
            }
            catch (InterruptedException ex) {
                throw new InterruptedNamingException("Interrupted while waiting for a connection");
            }
            if (n <= 0L) {
                continue;
            }
            n -= System.currentTimeMillis() - n2;
        }
        this.d("get(): after");
        return orCreateConnection;
    }
    
    private PooledConnection getOrCreateConnection(PooledConnectionFactory pooledConnectionFactory) throws NamingException {
        int size = this.conns.size();
        if (this.prefSize <= 0 || size >= this.prefSize) {
            int i = 0;
            while (i < size) {
                PooledConnection tryUse;
                if ((tryUse = ((ConnectionDesc)this.conns.get(i)).tryUse()) != null) {
                    this.d("get(): use ", tryUse);
                    this.td("Use ", tryUse);
                    return tryUse;
                }
                ++i;
            }
        }
        if (this.maxSize > 0 && size >= this.maxSize) {
            return null;
        }
        PooledConnection pooledConnection = pooledConnectionFactory.createPooledConnection(this);
        this.td("Create and use ", pooledConnection, pooledConnectionFactory);
        this.conns.add(new ConnectionDesc(pooledConnection, true));
        return pooledConnection;
    }
    
    @Override
    public synchronized boolean releasePooledConnection(PooledConnection pooledConnection) {
        ConnectionDesc connectionDesc;
        int index = this.conns.indexOf(connectionDesc = new ConnectionDesc(pooledConnection));
        this.d("release(): ", pooledConnection);
        if (index >= 0) {
            if (this.closed || (this.prefSize > 0 && this.conns.size() > this.prefSize)) {
                this.d("release(): closing ", pooledConnection);
                this.td("Close ", pooledConnection);
                this.conns.remove(connectionDesc);
                pooledConnection.closeConnection();
            }
            else {
                this.d("release(): release ", pooledConnection);
                this.td("Release ", pooledConnection);
                ((ConnectionDesc)this.conns.get(index)).release();
            }
            this.notifyAll();
            this.d("release(): notify");
            return true;
        }
        else {
            return false;
        }
    }
    
    @Override
    public synchronized boolean removePooledConnection(PooledConnection pooledConnection) {
        if (this.conns.remove(new ConnectionDesc(pooledConnection))) {
            this.d("remove(): ", pooledConnection);
            this.notifyAll();
            this.d("remove(): notify");
            this.td("Remove ", pooledConnection);
            if (this.conns.isEmpty()) {
                this.ref = null;
            }
            return true;
        }
        else {
            this.d("remove(): not found ", pooledConnection);
            return false;
        }
    }
    
    boolean expire(long n) {
        ArrayList list;
        synchronized (this) {
            list = new ArrayList(this.conns);
        }
        ArrayList<ConnectionDesc> list2 = new ArrayList();
        for (ConnectionDesc connectionDesc : list) {
            this.d("expire(): ", connectionDesc);
            if (connectionDesc.expire(n)) {
                list2.add(connectionDesc);
                this.td("expire(): Expired ", connectionDesc);
            }
            else {
                continue;
            }
        }
        synchronized (this) {
            this.conns.removeAll(list2);
            return this.conns.isEmpty();
        }
    }
    
    synchronized void close() {
        this.expire(System.currentTimeMillis());
        this.closed = true;
    }
    
    String getStats() {
        int i = 0;
        int j = 0;
        int k = 0;
        long lng = 0L;
        int size;
        synchronized (this) {
            size = this.conns.size();
            for (int l = 0; l < size; ++l) {
                ConnectionDesc connectionDesc = (ConnectionDesc)this.conns.get(l);
                lng += connectionDesc.getUseCount();
                switch (connectionDesc.getState()) {
                    case 0: {
                        ++j;
                        break;
                    }
                    case 1: {
                        ++i;
                        break;
                    }
                    case 2: {
                        ++k;
                        break;
                    }
                }
            }
        }
        return "size=" + size + "; use=" + lng + "; busy=" + j + "; idle=" + i + "; expired=" + k;
    }
    
    private void d(String str, Object obj) {
        if (Connections.debug) {
            this.d(str + obj);
        }
    }
    
    private void d(String str, int i) {
        if (Connections.debug) {
            this.d(str + i);
        }
    }
    
    private void d(String str) {
        if (Connections.debug) {
            System.err.println(this + "." + str + "; size: " + this.conns.size());
        }
    }
    
    private void td(String str, Object obj, Object obj2) {
        if (Connections.trace) {
            this.td(str + obj + "[" + obj2 + "]");
        }
    }
    
    private void td(String str, Object obj) {
        if (Connections.trace) {
            this.td(str + obj);
        }
    }
    
    private void td(String x) {
        if (Connections.trace) {
            System.err.println(x);
        }
    }
    
    static {
        debug = Pool.debug;
        trace = LdapPoolManager.trace;
    }
}
